package main

import (
	"fmt"
	"io"
	"log"
	"net/http"
	"os"
	"strings"
	"sync"

	"gitlab.com/gitlab-org/gitlab-metrics-exporter/internal/probe"
	"gitlab.com/gitlab-org/gitlab-metrics-exporter/internal/probe/mmap"
	"gitlab.com/gitlab-org/gitlab-metrics-exporter/internal/probe/self"
	"gitlab.com/gitlab-org/gitlab-metrics-exporter/internal/render"

	"gitlab.com/gitlab-org/labkit/metrics"

	"github.com/urfave/cli/v2"
)

const (
	metricsEndpoint  = "/metrics"
	metricsNamespace = "gitlab_metrics_exporter"

	probeSelf = "self"
	probeMmap = "mmap"

	defaultPort = "8082"
)

var (
	defaultProbes   = []string{probeSelf}
	availableProbes = []string{probeSelf, probeMmap}
)

type config struct {
	host   string
	port   string
	probes []probe.Probe
}

func main() {
	err := buildCliApp().Run(os.Args)
	if err != nil {
		log.Fatalf("server terminated: %v", err)
	}
}

func buildCliApp() *cli.App {
	return &cli.App{
		Usage: "A Prometheus exporter for GitLab application metrics",
		Flags: buildCliFlags(),
		Action: func(c *cli.Context) error {
			cfg, err := loadConfig(c)
			if err == nil {
				startServer(&cfg)
			}
			return err
		},
	}
}

func buildCliFlags() []cli.Flag {
	return []cli.Flag{
		&cli.StringFlag{
			Name:    "host",
			Usage:   "TCP server address host",
			EnvVars: []string{"GME_SERVER_HOST"},
		},
		&cli.StringFlag{
			Name:    "port",
			Value:   defaultPort,
			Usage:   "TCP server address port",
			EnvVars: []string{"GME_SERVER_PORT"},
		},
		&cli.StringSliceFlag{
			Name:    "probe",
			Value:   cli.NewStringSlice(defaultProbes...),
			Usage:   "probe to run when metrics are requested (available: " + strings.Join(availableProbes, ", ") + ")",
			EnvVars: []string{"GME_PROBES"},
			Aliases: []string{"p"},
		},
		&cli.StringFlag{
			Name:    "mmapdir",
			Usage:   "mmap probe: absolute path to prometheus-client-mmap metrics files",
			EnvVars: []string{"GME_MMAP_METRICS_DIR"},
		},
	}
}

func loadConfig(c *cli.Context) (config, error) {
	cfg := config{
		host: c.String("host"),
		port: c.String("port"),
	}

	probes, err := loadProbes(c)
	if err != nil {
		return cfg, err
	}
	cfg.probes = probes

	return cfg, nil
}

func loadProbes(c *cli.Context) ([]probe.Probe, error) {
	probeNames := c.StringSlice("probe")
	var probes []probe.Probe
	for _, probeName := range probeNames {
		if !isSupportedProbe(probeName) {
			return nil, fmt.Errorf("unsupported probe: %s", probeName)
		}

		switch probeName {
		case "self":
			probes = append(probes, self.NewProbe())
		case "mmap":
			metricsDir := c.String("mmapdir")
			if metricsDir == "" {
				return nil, fmt.Errorf("cannot be blank: --mmapdir")
			}
			probes = append(probes, mmap.NewDefaultProbe(metricsDir))
		}
	}

	return probes, nil
}

func isSupportedProbe(probe string) bool {
	for _, p := range availableProbes {
		if p == probe {
			return true
		}
	}
	return false
}

func startServer(cfg *config) {
	srvAddr := cfg.host + ":" + cfg.port
	log.Println("Serving /metrics on", srvAddr, "[ pid: ", os.Getpid(), "]")
	http.Handle(metricsEndpoint,
		metrics.NewHandlerFactory(metrics.WithLabels("path"), metrics.WithNamespace(metricsNamespace))(
			http.HandlerFunc(func(rw http.ResponseWriter, r *http.Request) {
				writeProbeResults(cfg.probes, rw)
			}),
			metrics.WithLabelValues(map[string]string{"path": metricsEndpoint}),
		))
	log.Fatal(http.ListenAndServe(srvAddr, nil))
}

func writeProbeResults(probes []probe.Probe, w io.Writer) {
	ch := make(chan probe.ProbeResult)

	wg := sync.WaitGroup{}

	for _, p := range probes {
		wg.Add(1)

		p_ := p
		go func() {
			defer wg.Done()
			if err := p_.Probe(ch); err != nil {
				log.Println(err)
			}
		}()
	}

	go func() {
		defer close(ch)
		wg.Wait()
	}()

	if err := render.NewRenderer().TextAsync(ch, w); err != nil {
		log.Println(err)
	}
}
